"""Common configuration constants
"""

PROJECTNAME = 'Products.PFGMasterSelect'

ADD_PERMISSIONS = {
    # -*- extra stuff goes here -*-
    'FormMasterSelectStringField': 'Products.PFGMasterSelect: Add FormMasterSelectStringField',
    'FormMasterMultiSelectStringField': 'Products.PFGMasterSelect: Add FormMasterMultiSelectStringField',
}



from zope.interface import Interface


class IFormMasterSelectStringField(Interface):
    """Description of the Example Type"""


class IFormMasterMultiSelectStringField(Interface):
    """Description of the Example Type"""



from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting
from plone.testing import z2
from zope.configuration import xmlconfig


class PFGMasterSelectLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import Products.PloneFormGen
        import Products.DataGridField
        import Products.MasterSelectWidget
        import Products.PFGMasterSelect
        xmlconfig.file('configure.zcml', Products.PloneFormGen, context=configurationContext)
        xmlconfig.file('configure.zcml', Products.DataGridField, context=configurationContext)
        xmlconfig.file('configure.zcml', Products.MasterSelectWidget, context=configurationContext)
        xmlconfig.file('configure.zcml', Products.PFGMasterSelect, context=configurationContext)

        # Install products that use an old-style initialize() function
        z2.installProduct(app, 'Products.PloneFormGen')
        z2.installProduct(app, 'Products.DataGridField')
        z2.installProduct(app, 'Products.MasterSelectWidget')
        z2.installProduct(app, 'Products.PFGMasterSelect')

    def tearDownZope(self, app):
        z2.uninstallProduct(app, 'Products.PFGMasterSelect')
        z2.uninstallProduct(app, 'Products.MasterSelectWidget')
        z2.uninstallProduct(app, 'Products.DataGridField')
        z2.uninstallProduct(app, 'Products.PloneFormGen')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'Products.PloneFormGen:default')
        applyProfile(portal, 'Products.DataGridField:default')
        applyProfile(portal, 'Products.MasterSelectWidget:default')
        applyProfile(portal, 'Products.PFGMasterSelect:default')

MASTER_SELECT_FIXTURE = PFGMasterSelectLayer()
MASTER_SELECT_INTEGRATION_TESTING = IntegrationTesting(
    bases=(MASTER_SELECT_FIXTURE,),
    name="PFGMasterSelectLayer:Integration"
)
MASTER_SELECT_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(MASTER_SELECT_FIXTURE, z2.ZSERVER_FIXTURE),
    name="PFGMasterSelectLayer:Functional"
)



from Products.CMFCore.Expression import Expression
from Products.validation import validation
from Products.validation.interfaces.IValidator import IValidator
from zope.interface import implements


class SlaveConfigValidator:
    """Validate as True when having at least one DataGrid item.
    """

    implements(IValidator)

    def __init__(self, name, title='', description=''):
        self.name = name
        self.title = title or name
        self.description = description

    def __call__(self, value, *args, **kwargs):
        errors = []
        for field in value:
            for method in ('toggle_method', 'vocab_method',):
                if method in field and field[method]:
                    try:
                        expr = Expression('python: ' + field[method])
                    except Exception as e:
                        errors.append('Invalid %s "%s" for %s/%s'
                                      % (method, field[method], field['name'], field['action']))

        return True if not errors else '; '.join(errors)


isSlaveConfigValid = SlaveConfigValidator('isSlaveConfigValid')
validation.register(isSlaveConfigValid)



import config
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_hasattr
from Products.Archetypes.public import listTypes


def importVarious(context):
    """
    Final PFGMasterSelect import steps.
    """

    # Only run step if a flag file is present (e.g. not an extension profile)
    if context.readDataFile('pfgmasterselect-various.txt') is None:
        return

    site = context.getSite()

    #######################
    # Both PloneFormGen and the target field provider are going to have
    # to be installed first. So, let's check.

    portal_skins = getToolByName(site, 'portal_skins')
    #assert safe_hasattr(portal_skins, 'MasterSelectWidget'), "MasterSelectWidget must be installed prior to installing this product."
    assert safe_hasattr(portal_skins, 'PloneFormGen'), "PloneFormGen must be installed prior to installing this product."

    #######################
    # Here's the code specific to making this visible to PloneFormGen

    classes = listTypes(config.PROJECTNAME)
    myTypes = [item['name'] for item in classes]
    portal_types = getToolByName(site, 'portal_types')
    for typeName in ('FormFolder', 'FieldsetFolder'):
        ptType = portal_types.getTypeInfo(typeName)
        ffact = list(ptType.allowed_content_types)
        ffact += myTypes
        ptType.manage_changeProperties(allowed_content_types=ffact)



from Products.ATContentTypes.content import schemata
from Products.Archetypes import atapi
from Products.CMFCore.Expression import Expression
from Products.CMFCore.permissions import View
from Products.DataGridField import DataGridField
from Products.DataGridField import DataGridWidget
from Products.DataGridField import SelectColumn
from Products.MasterSelectWidget.MasterMultiSelectWidget import MasterMultiSelectWidget
from Products.MasterSelectWidget.MasterSelectWidget import MasterSelectWidget
from Products.PloneFormGen.content.fieldsBase import BaseFormField, BaseFieldSchemaStringDefault, vocabularyField, \
    vocabularyOverrideField, LinesVocabularyField, StringVocabularyField
from zope.interface import implements
from .interfaces import IFormMasterSelectStringField, IFormMasterMultiSelectStringField
from Products.PFGMasterSelect.config import PROJECTNAME


class FormMasterSelectFieldMixin:

    def setSlave_fields(self, slave_fields):
        self.Schema()['slave_fields'].set(self, slave_fields)

        def compile_methods(field):
            field = field.copy()
            if 'toggle_method' in field:
                toggle_method = field['toggle_method']
                field['toggle_method'] = Expression('python:' + toggle_method) if toggle_method else None
            if 'vocab_method' in field:
                vocab_method = field['vocab_method']
                field['vocab_method'] = Expression('python:' + vocab_method) if vocab_method else None
            return field

        self.fgField.widget.slave_fields = [compile_methods(field)
                                            for field in slave_fields
                                            if field.get('toggle_method')
                                            or field.get('vocab_method')
                                            or field.get('hide_values')]

    def getActionVocab(self):
        return atapi.DisplayList((
            ('hide', 'Hide'),
            ('show', 'Show'),
            ('enable', 'Enable'),
            ('disable', 'Disable'),
            ('value', 'Value'),
            ('vocabulary', 'Vocabulary'),
        ))

    def getNameVocab(self):
        folder = self.aq_parent
        fieldNames = [field.getName() for field in folder.fgFields()]
        return atapi.DisplayList(zip(fieldNames, fieldNames))


FormMasterSelectFieldSchema = BaseFieldSchemaStringDefault.copy() + atapi.Schema((
    vocabularyField,
    vocabularyOverrideField,
    DataGridField('slave_fields',
                  allow_insert=True,
                  allow_delete=True,
                  allow_reorder=True,
                  columns=('name', 'action', 'vocab_method', 'toggle_method', 'hide_values'),
                  validators=('isColumnFilled', 'isSlaveConfigValid',),
                  widget=DataGridWidget(label='Slave Fields',
                                        description='Configure actions applied on other fields of the Form Folder '
                                                    'when changing this fields state. vocab_method and toggle_method '
                                                    'expect python expressions evaluating to a boolean. The new value '
                                                    'of this field is available as "value" of type string in '
                                                    'vocab_method and toggle_method.',
                                        columns={
                                              'name': SelectColumn('Name', 'getNameVocab'),
                                              'action': SelectColumn('Action', 'getActionVocab'),
                                        }),
                  ),
))

schemata.finalizeATCTSchema(FormMasterSelectFieldSchema, moveDiscussion=False)


class FormMasterSelectStringField(BaseFormField, FormMasterSelectFieldMixin):

    implements(IFormMasterSelectStringField)

    meta_type = "FormMasterSelectStringField"
    schema = FormMasterSelectFieldSchema

    def __init__(self, oid, **kwargs):
        BaseFormField.__init__(self, oid, **kwargs)

        self.fgField = StringVocabularyField('fg_masterselect',
                                             searchable=False,
                                             required=False,
                                             vocabulary=None,
                                             write_permission=View,
                                             widget=MasterSelectWidget(slave_fields=[]))


FormMasterMultiSelectFieldSchema = BaseFieldSchemaStringDefault.copy() + atapi.Schema((
    vocabularyField,
    vocabularyOverrideField,
    DataGridField('slave_fields',
                  allow_insert=True,
                  allow_delete=True,
                  allow_reorder=True,
                  columns=('name', 'action', 'vocab_method', 'toggle_method', 'hide_values'),
                  validators=('isColumnFilled', 'isSlaveConfigValid',),
                  widget=DataGridWidget(label='Slave Fields',
                                        description='Configure actions applied on other fields of the Form Folder '
                                                    'when changing this fields state. vocab_method and toggle_method '
                                                    'expect python expressions evaluating to a boolean. The selected '
                                                    'values of this field are available as "values" of type list of '
                                                    'strings in vocab_method and toggle_method.',
                                        columns={
                                         'name': SelectColumn('Name', 'getNameVocab'),
                                         'action': SelectColumn('Action', 'getActionVocab'),
                                        }),
                  ),
))

schemata.finalizeATCTSchema(FormMasterMultiSelectFieldSchema, moveDiscussion=False)


class FormMasterMultiSelectStringField(BaseFormField, FormMasterSelectFieldMixin):

    implements(IFormMasterMultiSelectStringField)

    meta_type = "FormMasterMultiSelectStringField"
    schema = FormMasterMultiSelectFieldSchema

    def __init__(self, oid, **kwargs):
        BaseFormField.__init__(self, oid, **kwargs)

        self.fgField = LinesVocabularyField('fg_masterselect',
                                            searchable=False,
                                            required=False,
                                            vocabulary=None,
                                            multiValued=True,
                                            write_permission=View,
                                            widget=MasterMultiSelectWidget(slave_fields=[]))


atapi.registerType(FormMasterSelectStringField, PROJECTNAME)
atapi.registerType(FormMasterMultiSelectStringField, PROJECTNAME)



import unittest

from Products.CMFCore import Expression
from Products.PFGMasterSelect.browser import JSONValuesForAction
from mock import MagicMock, ANY
from plone.testing.z2 import Browser

import transaction
from Products.PFGMasterSelect.testing import MASTER_SELECT_FUNCTIONAL_TESTING, MASTER_SELECT_INTEGRATION_TESTING
from plone.app.testing import TEST_USER_ID, setRoles



class TestJSONValuesForAction(unittest.TestCase):

    layer = MASTER_SELECT_INTEGRATION_TESTING

    def setUp(self):
        self.context = MagicMock()
        self.context.findFieldObjectByName.return_value = None
        self.context.provides.return_value = True
        self.request = MagicMock(field=None, slave=None, value=None, action=None)
        self.obj = JSONValuesForAction(self.context, self.request)

    def tearDown(self):
        del self.context
        del self.request
        del self.obj

    def test_getSlaves_none_fieldname(self):
        with self.assertRaises(AttributeError):
            self.obj.getSlaves(None)

    def test_getSlaves_empty_fieldname(self):
        with self.assertRaises(AttributeError):
            self.obj.getSlaves('')

    def test_getSlaves_unknown_fieldname(self):
        with self.assertRaises(AttributeError):
            self.obj.getSlaves('unknnown')

    def test_getSlaves_known_fieldname(self):
        SLAVE_FIELDS = (1, 2, 3,)
        field_object = MagicMock()
        field_object.fgField.widget.slave_fields = SLAVE_FIELDS
        context = MagicMock()
        context.findFieldObjectByName.return_value = field_object
        request = dict(field=None, slave=None, value=None, action=None)
        obj = JSONValuesForAction(context, request)
        self.assertEquals(SLAVE_FIELDS, obj.getSlaves('known'))

    def test_evaluate_method_none_method(self):
        with self.assertRaises(AssertionError):
            self.obj.evaluate_method(None, [])

    def test_evaluate_method_single(self):
        VALUE = '2'
        expr = Expression.Expression('python: value')
        args = VALUE
        self.assertEquals(VALUE, self.obj.evaluate_method(expr, args))

    def test_evaluate_method_multi(self):
        VALUES_SELECTED = ('1', '2',)
        VALUES_NOT_SELECTED = ('4',)
        expr = Expression.Expression('python: sum(map(int, values))')
        args = [dict(selected=True, val=v) for v in VALUES_SELECTED] \
               + [dict(selected=False, val=v) for v in VALUES_NOT_SELECTED]
        self.assertEquals(3, self.obj.evaluate_method(expr, args))

    def test_argument_format_expectation_single(self):
        self.assertEquals([42], self.obj.extractArguments(None, '42'))

    def test_argument_format_expectation_multi(self):
        self.assertEquals([dict(selected=True, val='3'), dict(selected=False, val='7')],
                          self.obj.extractArguments(None, '[{"selected":true,"val":"3"},{"selected":false, "val":"7"}]'))

    def test_extracted_arguments_passed_directly_to_compute(self):
        SLAVES = [dict(name='a', action='hide')]
        VALUE = '"test43"'
        self.obj.action = 'hide'
        self.obj.value = VALUE
        self.obj.slaveid = SLAVES[0]['name']
        self.obj.getSlaves = MagicMock(return_value=SLAVES)
        self.obj.extractArguments = MagicMock(return_value=VALUE)
        self.obj.computeJSONValues = MagicMock()
        self.obj()

        self.obj.computeJSONValues.assert_called_with(ANY, VALUE)


class Test(unittest.TestCase):

    layer = MASTER_SELECT_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.app = self.layer['app']
        self.request = self.layer['request']

        portal = self.portal

        setRoles(self.portal, TEST_USER_ID, ['Manager',])

        portal.invokeFactory('FormFolder', 'formfolder')
        self.form = portal.formfolder

        self.form.invokeFactory('FormMasterSelectStringField', 'master')
        self.master = self.form.master

        self.form.invokeFactory('FormMasterSelectStringField', 'slave')
        slave = self.form.slave

        self.master.setFgVocabulary(['1', '2'])
        self.master.setSlave_fields([dict(name='slave', action='hide', vocab_method='', toggle_method='value in ("1")', hide_values='2')])

        transaction.commit()

        self.browser = Browser(self.app)
        self.browser.handleErrors = False

    def test(self):
        browser = self.browser
        browser.open(self.form.absolute_url() + '/@@masterselect-jsonvalue-toggle?field=master&slave=slave&action=hide&value=1')
        self.assertTrue('"action": "show"' in browser.contents)
        self.assertTrue('"toggle": false' in browser.contents)



import json

import logging

import Products
from App.Common import aq_base
from Products.Archetypes import DisplayList
from Products.CMFCore import Expression
from Products.CMFCore.Expression import getExprContext
from Products.Five import BrowserView
from Products.MasterSelectWidget.browser import JSONValuesForAction, SetupSlaves
from Products.PageTemplates.Expressions import getEngine
from Products.PloneFormGen.content.form import FormFolder
from zope.i18n import translate


logger = logging.getLogger('PFGMasterSelect')


class Migration(BrowserView):

    def fix_write_permissions(self):
        from plone.protect.interfaces import IDisableCSRFProtection
        from zope.interface import alsoProvides
        alsoProvides(self.request, IDisableCSRFProtection)

        catalog = self.context.portal_catalog

        for brain in catalog(portal_type=('FormMasterSelectStringField', 'FormMasterMultiSelectStringField',)):
            obj = brain.getObject()

            from Products.CMFCore.permissions import View
            obj.fgField.write_permission = View
            obj._p_changed = True

            logger.warning('Fixed write permission of object %s' % '/'.join(obj.getPhysicalPath()))

    def migrate_to_v02(self):
        from plone.protect.interfaces import IDisableCSRFProtection
        from zope.interface import alsoProvides
        alsoProvides(self.request, IDisableCSRFProtection)

        catalog = self.context.portal_catalog

        for brain in catalog(portal_type=('FormMasterSelectStringField', 'FormMasterMultiSelectStringField',)):
            obj = brain.getObject()

            logger.warning('Updating slave field configuration of object %s' % '/'.join(obj.getPhysicalPath()))

            slave_fields = obj.getSlave_fields()
            obj.setSlave_fields(slave_fields)
            obj._p_changed = True


class SetupSlaves(SetupSlaves):
    pass


class JSONValuesForAction(JSONValuesForAction):

    def evaluate_method(self, method, args):
        assert method
        assert isinstance(method, Expression.Expression)

        # Assumptions:
        # There are two possibilities: This is called on a (a) single select field or (b) multi select field.
        # In case (a) there should always be exactly one value, hence len(args) == 1, and it should not be a
        # dictionary. On the other hand in case (b) there can be an arbitrary count of elements selected and all
        # elements are stored in dictionaries.
        if len(args) == 0 or isinstance(args[0], dict):
            data = dict(values=[str(arg['val']) for arg in args if arg['selected']])
        else:
            assert len(args) == 1
            data = dict(value=str(args[0]))

        econtext = getEngine().getContext(data)

        try:
            result = method(econtext)
        except Exception as e:
            logger.error('%s for "%s %s" of %s/%s crashed: %s',
                         'vocab_method' if self.action in ('vocabulary', 'value') else 'toggle_method',
                         self.action,
                         self.slaveid,
                         '/'.join(self.context.getPhysicalPath()),
                         self.field,
                         e)
            result = None
        return result

    def getSlaves(self, fieldname):
        #assert isinstance(self.context, FormFolder)

        field_object = self.context.findFieldObjectByName(fieldname)
        slave_fields = field_object.fgField.widget.slave_fields or ()
        return slave_fields

    def __call__(self):
        #assert isinstance(aq_base(self.context), FormFolder)
        return Products.MasterSelectWidget.browser.JSONValuesForAction.__call__(self)


class JSONValuesForVocabularyChange(JSONValuesForAction):

    def computeJSONValues(self, slave, args):
        method = slave['vocab_method']
        vocabulary = self.evaluate_method(method, args) if method else ()
        if not isinstance(vocabulary, (list, tuple,)):
            vocabulary = (vocabulary,)
        vocabulary = [str(item) for item in vocabulary]
        vocabulary = DisplayList(zip(vocabulary, vocabulary))

        return json.dumps([
                dict(
                    value=item,
                    label=translate(vocabulary.getValue(item), context=self.request)
                ) for item in vocabulary
            ])


class JSONValuesForValueUpdate(JSONValuesForAction):

    def computeJSONValues(self, slave, args):
        method = slave['vocab_method']
        value = self.evaluate_method(method, args) if method else None
        return json.dumps(translate(value, context=self.request))


class JSONValuesForToggle(JSONValuesForAction):

    def computeJSONValues(self, slave, args):
        method = slave['toggle_method']
        if method:
            toggle = self.evaluate_method(method, args)
        else:
            hide_values = slave.get('hide_values')
            hide_values = [s.strip() for s in hide_values.split(',')]
            toggle = str(args[0]) in hide_values

        action = self.action
        if action in ['disable', 'hide']:
            toggle = not toggle
            action = action == 'disable' and 'enable' or 'show'
        json_toggle = json.dumps({'toggle': toggle, 'action': action})

        return json_toggle
